###
# Affine Geometry with Sage
#
# by Jaume Aguad'e (Jaume.Aguade@uab.cat)
#
# A sage companion to Agust'i Revent'os
# "Affine Maps, Euclidean Motions and Quadrics"
# Springer Undergraduat Mathematics Series 2011
# 
# Version 1.1
#
# Sage version 6.9
#
# Read the user manual "Affine_Geometry_with_Sage.pdf"
###
#
#
# Chapter 1. Affine Space
#
import copy
###
### Adapting to version 6.9
# Image of a subspace under a matrix
# V is a subspace of E, A is a matrix acting on E. The function returns the subspace A(V)
def ImageUnderMatrix(E,V,A):
  return E.subspace([A*e for e in V.basis()])
###
### Some auxiliar functions
###
###
# this function inputs an element in AA (or a list or a matrix)
# and returns (if possible) an element x in SR
# such that AA(x)==a
#
def Symbolize(a,size='element'):
  if size=='element':
    b=AA(a)
    p=b.minpoly()
    if p.degree()==1:
      return a
    z=var('z')
    S=solve(p(z)==0,z)
    if len(S)==1:
      raise NameError("cannot symbolize")
    H=[S[i].rhs() for i in range(len(S))]
    HQ=[QQbar(h) for h in H]
    HA=[AA(h) for h in HQ if h in AA]
    for i in range(len(H)):
      if HA[i]==b:
        return H[i]
  if size=='list':
    return [Symbolize(a[i]) for i in range(len(a))]
  if size=='matrix':
    return matrix(SR,[Symbolize(a.row(i),size='list') for i in range(a.nrows())])
  if size=='polynomial':
    M=a.monomials()
    C=a.coefficients()
    return sum([SR(Symbolize(C[i]))*SR(str(M[i])) for i in range(len(M))])
###
# Extend to a basis
# B is a set of linearly independent vectors in V.
# V is a subspace of E.
# This function returns a basis of V which starts with B.
#
def extend(B,V,E):
  r=len(B)
  if r==0:
    return [E(V.basis()[i]) for i in range(V.dimension())]
  Bprime=[B[i] for i in range(r) if B[i] in V]
  Vprime=E.subspace(B)
  if not B==Bprime:
    raise NameError("vectors in partial basis must be in subspace")
  if not Vprime.dimension()==r:
    raise NameError("vectors must be linearly independent")
  X=copy.copy(B)
  while len(X)<V.dimension():
    for i in range(V.dimension()):
      if not V.basis()[i] in E.subspace(X):
        X.append(E(V.basis()[i]))
        break 
  return X
###
### Funcions for Chapter 1 in the book: 
### Affine spaces, varieties, frames and equations
###
#
def AffineSpace(K,d,metric=0):
  if metric==0:
    V=VectorSpace(K,d,inner_product_matrix=identity_matrix(K,d))
  else:
    V=VectorSpace(K,d,inner_product_matrix=metric)
  return V,V
#
def Arrow(P,Q):
  return Q-P
#
def Variety(P,E):
  return [P]+E.basis()
#
def BasePoint(X):
  return X[0]
#
def Direction(X):
  V=X[0].parent()
  L=X[1:len(X)]
  return V.subspace(L)
#
def PointAsVariety(P):
  return Variety(P,parent(P).subspace([]))
#  
def AffineDimension(X):
  return Direction(X).dimension()
#
def AffineCodimension(X):
  V=parent(X[0]); d=V.dimension()
  r=AffineDimension(X)
  return d-r
#  
def Belongs(P,X):
  V=Direction(X)
  return Arrow(X[0],P) in V
#  
def Contained(X,Y):
  F=Direction(Y); G=Direction(X)
  return G.is_subspace(F) and Belongs(BasePoint(X),Y)
#
def EqualityOfVarieties(X,Y):
  return Contained(X,Y) and AffineDimension(X)==AffineDimension(Y)
#    
def Parallel(X,Y):
  F=Direction(Y); G=Direction(X)
  return G.is_subspace(F) or F.is_subspace(G)
#    
def AffineSum(S):
  if len(S)==0:
    raise NameError("nothing to add")
  if len(S)==1:
    return S[0]
  X=AffineSum(S[1:]); Y=S[0]
  F=Direction(Y); G=Direction(X)
  E=parent(BasePoint(X))
  W=F+G+E.subspace([BasePoint(Y)-BasePoint(X)])
  return Variety(BasePoint(X),W)
#
def Meet(X,Y):
  P=BasePoint(X); V=Direction(X)
  Q=BasePoint(Y); W=Direction(Y)
  M=matrix(V.basis()+W.basis())
  if (P-Q) in V+W:
    h=M.solve_left(P-Q)
    u1=[h[i] for i in range(V.dimension())] + [0 for i in [V.dimension()..len(h)-1]]
    uV=vector(u1)*M
    return True,Variety(P-uV,V.intersection(W))
  else:
    return False,False
#  
def Frame(P,B):
  V=parent(P)
  W=V.subspace(B)
  if not V==W or not len(B)==V.dimension():
    raise NameError("Not an affine frame")
  return [P]+B
#
def CanonicalFrame(A):
  return Frame(A(0),A.basis())
#  
def FrameAsMatrix(R):
  P=R[0]      
  d=parent(P).dimension()
  S=[]; T=[]
  for i in [1..d]:
    S.append([R[i][j] for j in range(d)])
    S[i-1].append(0)
    T.append(P[i-1])
  T.append(1)
  return matrix(S+[T]).transpose()
#  
def ChangeOfFrameMatrix(R,S):
  return FrameAsMatrix(S)^(-1)*FrameAsMatrix(R)
#  
def AffineCoordinates(P,Ref):
  Q=Ref[0]
  S=Ref[1:len(Ref)]
  Z=parent(Q).subspace_with_basis(S)
  return Z.coordinate_vector(P-Q)
#  
def ChangeOfFramePoint(P,R,S):
  V=parent(P); d=V.dimension()
  M=ChangeOfFrameMatrix(R,S)
  X=FrameAsMatrix(Frame(P,V.basis()))
  Y=M*X
  return V([Y[i,d] for i in range(d)])
#  
def ChangeOfFrameVariety(X,Ref):
  V=parent(X[0]); d=V.dimension()
  C=CanonicalFrame(V)
  M=FrameAsMatrix(Ref)
  h=matrix([[M[i,j] for j in range(d)] for i in range(d)])
  f=linear_transformation(V,V,h.transpose(),side='left')
  return Variety(ChangeOfFramePoint(BasePoint(X),C,Ref),f(Direction(X)))  
#  
def Hyperplane(f,A):
  if not f.total_degree()==1:
    raise NameError("polynomial must be of degree one")
  d=A.dimension(); Poly=parent(f)
  for i in range(d):
    if not f.coefficient(Poly.gens()[i])==0:
      a=f.coefficient(Poly.gens()[i])
      b=i
      break 
  P=A(-(f.constant_coefficient()/a)*(A.basis()[b]))
  C=[]
  for i in range(d):
    C.append(f.coefficient(Poly.gens()[i]))
  phi=linear_transformation(A,VectorSpace(base_field(A),1),matrix(C).transpose())
  return Variety(P,kernel(phi))
#
def VarietyByEquations(S,A):
  if len(S)==0:
    return Variety(A(0),A)
  if len(S)==1:
    return Hyperplane(S[0],A)
  else:
    a,X=Meet(Hyperplane(S[0],A),VarietyByEquations(S[1:len(S)],A))
    if not a:
      raise NameError("variety is empty")
    return X
#
def CartesianEquations(X,Poly):
  F=Direction(X); r=F.dimension()
  P=BasePoint(X)
  V=parent(P); K=V.base_field()
  d=V.dimension()
  if r==d:
    raise NameError("variety is the full affine space")
  U,pi,lift=V.quotient_abstract(F); G=lift(U)
  B=F.basis()+G.basis()
  M=(matrix([V.coordinate_vector(B[i]) for i in range(d)]))^(-1)
  S=[]
  for i in [r..d-1]:
    f=sum([M[j,i]*Poly.gens()[j] for j in range(d)])
    S.append(f-f([P[i] for i in range(d)]))
  return S
# 
def Barycenter(S):
  l=len(S)
  if l==0:
    raise NameError("no points given")
  if not parent(S[0]).base_field().characteristic()==0:
    raise NameError("Barycenter requires characteristic zero")
  return sum([(1/l)*S[i] for i in range(l)])
#    
def Ratio(A,B,C):
  V=parent(A)
  if A==C:
    raise NameError("first and last point must be different")
  if not Belongs(B,AffineSum([PointAsVariety(A),PointAsVariety(C)])):
    raise NameError("points are not alligned")
  X=B-A; Y=C-A
  for i in range(V.dimension()):
    if not V.coordinates(Y)[i]==0:
      J=i
      break
  return V.coordinates(X)[J]/V.coordinates(Y)[J];
###
### Functions for Chapter 2 in the book:
### Affinities, fixed points, fixed lines
###
def Affinity(A,Q,h):
  def f(P):
    return Q+h(P)
  return f
#
def AffinityAsMatrix(f):
  Q=f(0); A=parent(Q); d=A.dimension()
  S=[]; T=[]
  for i in range(d):
    v=f(A.basis()[i])-Q
    S.append([A.coordinate_vector(v)[j] for j in range(d)])
    S[i].append(0)
    T.append(Q[i])
  T.append(1)
  return matrix(S+[T]).transpose()
#
def IsBijective(f):
  M=AffinityAsMatrix(f)
  if M.determinant()==0:
    return False
  else:
    return True
# 
def EquationsOfAffinity(f,Poly):
  Q=f(0); A=parent(Q); d=A.dimension()
  M=AffinityAsMatrix(f)
  S=[]
  for i in range(d):
    S.append(sum([M[i,j]*Poly.gens()[j] for j in range(d)])+M[i,d])
  return S
#
def AffinityByEquations(S,A):
  d=A.dimension()
  if not len(S)==d:
    raise NameError("number of equations is wrong")
  for i in range(d):
    if not S[i].total_degree()<=1:
      raise NameError("equations must be of degree one")
  Poly=parent(S[0])
  Q=A([S[i].constant_coefficient() for i in range(d)])
  v=[]
  for i in range(d):
    v.append(A([S[j].coefficient(Poly.gens()[i]) for j in range(d)]))
  h=Hom(A,A)(matrix(v))
  return Affinity(A,Q,h)
#
def LinearPartOfAffinity(f):
  Q=f(0); A=parent(Q); d=A.dimension()
  h=Hom(A,A)(matrix([f(A.basis()[i])-Q for i in range(d)]))
  return h
#
def MatrixAsAffinity(M,A):
  d=A.dimension()
  for i in range(d):
    if not M[d,i]==0:
      raise NameError("not the matrix of an affinity")
  if not M[d,d]==1:
    raise NameError("not the matrix of an affinity")
  h=Hom(A,A)([[M[i,j] for i in range(d)] for j in range(d)])
  Q=A([M[i,d] for i in range(d)])
  return Affinity(A,Q,h)
#    
def EqualityOfAffinities(f,g):
  return AffinityAsMatrix(f)==AffinityAsMatrix(g)
#
def InverseOfAffinity(f):
  if not IsBijective(f):
    raise NameError("Affinity must be bijective")
  M=AffinityAsMatrix(f)
  return MatrixAsAffinity(M^(-1),parent(f(0)))
#
def CompositionOfAffinities(f,g):
  A=parent(f(0))
  M=AffinityAsMatrix(f)
  N=AffinityAsMatrix(g)
  return MatrixAsAffinity(M*N,A)
#
def Conjugate(f,g):
  if not IsBijective(g):
    raise NameError("Second affinity must be bijective")
  M=AffinityAsMatrix(f)
  N=AffinityAsMatrix(g)
  return MatrixAsAffinity(N^(-1)*M*N,parent(f(0)))  
#
def FixedPoints(f):
  Q=f(0); A=parent(Q); d=A.dimension()
  h=LinearPartOfAffinity(f)
  Id=Hom(A,A)(identity_matrix(d))
  try:
    X=(h-Id).preimage_representative(-Q)
    return Variety(X,(h-Id).kernel());
  except ValueError:
    raise NameError("There are no fixed points")
#
def ImageOfVariety(f,X):
  h=LinearPartOfAffinity(f)
  P=BasePoint(X)
  W=Direction(X)
  return Variety(f(P),h(W))
#    
def IsInvariant(X,f):
  Y=ImageOfVariety(f,X)
  return Contained(Y,X)
#
def SparseInvariantLines(f):
  E=parent(f(0)).change_ring(AA); d=E.dimension()
  h=LinearPartOfAffinity(f)
  try:
    Fix=FixedPoints(f)
  except NameError:
    raise NameError("SparseInvariantLines requires that affinity has a single fixed point")
  if not AffineDimension(Fix)==0:
    raise NameError("SparseInvariantLines requires that affinity has a single fixed point")
  EV=h.matrix().change_ring(QQbar).eigenvectors_left()
  Aevect=[e for e in EV if e[0] in AA]
  too_many=[e for e in Aevect if len(e[1])>1]
  if not too_many==[]:
    print 'there are infinitely many invariant lines'
  vectors=[E(e[1][0]) for e in Aevect]
  return [Variety(Fix[0],E.subspace([v])) for v in vectors]
#  
def SpecialInvariantLines(f):
  try:
    Fix=FixedPoints(f)
    fixed=True
  except NameError:
    fixed=False
  h=LinearPartOfAffinity(f)
  E=h.domain(); d=E.dimension()
  b=f(0)
  Id=identity_matrix(d)
  J=Hom(E,E)(2*Id)
  if (h-Hom(E,E)(Id)).is_invertible():
    raise NameError("there are no special invariant lines")
  else:
    T=h^2-J*h+Hom(E,E)(Id)
    try:
      X=T.preimage_representative((Hom(E,E)(Id)-h)(b))
      W=Variety(X,T.kernel())
      if fixed:
	if AffineDimension(W)==AffineDimension(Fix):
	  raise NameError("there are no special invariant lines")
      return W
    except ValueError:
      print "there are no special invariant lines"
# 
def AffineTranslation(v):
  E=parent(v);
  return Affinity(E,v,Hom(E,E).identity())
#  
def IsTranslation(f):
  h=LinearPartOfAffinity(f)
  v=f(0); E=parent(v)
  if h==Hom(E,E).identity():
    return True,v
  else:
    return False,False
#
def RandomTranslation(A):
  d=A.dimension()
  S=[]
  for i in range(d):
    S.append(ZZ.random_element())
  return AffineTranslation(A(S))
#
def AffineHomothecy(r,P):
  if r in [0,1]:
    raise NameError("ratio must be different to 0, 1")
  E=parent(P); d=E.dimension()
  h=Hom(E,E)(r*identity_matrix(d))
  return Affinity(E,P,h)
#  
def IsHomothecy(f):
  h=LinearPartOfAffinity(f)
  v=f(0); E=parent(v)
  p=h.matrix().minpoly()
  if not p.degree()==1:
    return False,False,False
  r=-p(0)
  return True,r,f(0)
#  
def RandomHomothecy(A):
  s=0
  while s in [0,1]:
    s=ZZ.random_element();
  d=A.dimension()
  S=[]
  for i in [1..d]:
    S.append(ZZ.random_element())
  return AffineHomothecy(s,A(S))
#
def Asplitvector(u,F,G):
  E=parent(u)
  M=matrix(F.basis()+G.basis())
  h=Hom(E,E)(M^(-1))
  v=h(u)
  w=sum([v[i]*F.basis()[i] for i in range(F.dimension())])
  return w
#
def Symmetry(X,G):
  P=BasePoint(X)
  E=parent(P); d=E.dimension()
  F=Direction(X)
  if not F+G==E or not F.dimension()+G.dimension()==E.dimension():
    raise NameError("second argument must be a complement to the direction subspace of the first argument")
  S=[]
  for i in range(d):
    S.append(2*Asplitvector(E.basis()[i],F,G)-E.basis()[i])
  h=Hom(E,E)(S)
  return Affinity(E,P-h(P),h)
#    
def IsSymmetry(f):
  E=parent(f(0))
  M=AffinityAsMatrix(f)
  h=LinearPartOfAffinity(f)
  G=(h+Hom(E,E).identity()).kernel()
  if M^2==parent(M).identity_matrix():
    return True,FixedPoints(f),G
  else:
    return False,False,False
#
def RandomSymmetry(A,dim):
  d=A.dimension()
  if dim > d:
    raise NameError("second argument cannot be greater than the dimension of the affine space")
  M=random_matrix(ZZ,d,algorithm='diagonalizable',eigenvalues=[1,-1],dimensions=[dim,d-dim]) 
  b=random_matrix(ZZ,1,d-dim)*matrix((M+identity_matrix(d)).kernel().basis())
  return Affinity(A,A([b[0,i] for i in range(d)]),Hom(A,A)(M))
#   
def AffineProjection(X,G):
  P=BasePoint(X)
  E=parent(P)
  F=Direction(X)
  if not F+G==E or not F.dimension()+G.dimension()==E.dimension(): 
    raise NameError("second argument must be a complement to the direction subspace of the first argument")
  S=[]
  M=matrix(F.basis()+G.basis())
  for i in range(F.dimension()):
    S.append([M[i,j] for j in range(E.dimension())])
  for i in [F.dimension()..E.dimension()-1]:
    S.append([0 for j in range(E.dimension())])
  h=Hom(E,E)(M^(-1)*matrix(S))
  return Affinity(E,P-h(P),h)
#
def IsAffineProjection(f):
  E=parent(f(0))
  M=AffinityAsMatrix(f)
  h=LinearPartOfAffinity(f)
  if M^2==M:
    return True,FixedPoints(f),h.kernel()
  else:
    return false,False,False
#
def RandomAffineProjection(A,dim):
  d=A.dimension()
  if dim > d:
    raise NameError("second argument cannot be greater than the dimension of the affine space")
  M=random_matrix(ZZ,d,algorithm='diagonalizable',eigenvalues=[1,0],dimensions=[dim,d-dim]) 
  b=random_matrix(ZZ,1,d-dim)*matrix(M.kernel().basis())
  return Affinity(A,A([b[0,i] for i in range(d)]),Hom(A,A)(M))
#
###
### Function for Chapters 3 and 4 in the book:
### Classification of affinities
###
###
# Computation of the invariance level of an affinity.
# Read the doc for explanation of this function.
#
def InvarianceLevel(f):
  v=f(0)
  E=parent(v); d=E.dimension()
  h=LinearPartOfAffinity(f)-Hom(E,E)(identity_matrix(d))
  inimage=False
  r=0
  while not inimage:
    inimage=((h^r)(v) in (h^(r+1)).image())
    r=r+1
  return r-1
###
# height of a vector
# M is a matrix and u is a vector killed by some (M-evalue*Id)^N 
# the function returns the minimum N.
# Default value for the eigenvalue eval is 1
#
def JordanHeight(u,M,evalue=1):
  E=parent(u); d=E.dimension(); K=E.base_field()
  Diag=matrix(K,d,d,evalue)
  return min([i for i in range(d+1) if u in ((M-Diag)^i).right_kernel()])
###
# M is a matrix which contains a Jordan block of size r and eigenvalue 1.
# The function returns a matrix similar to M with this Jordan block 
# of size r raised to the top.
#
def raise_to_top(M,r):
  s=M.ncols()
  Id=identity_matrix(s)
  S=[j for j in range(s) if M.column(j)==Id.column(j)]
  for j in S:
    h=max([i for i in [0..j] if M.submatrix(j-i,j-i,i+1,i+1)==jordan_block(1,i+1).transpose()])+1
    if h==r:
      pos=j
      break
  Mtop=M.submatrix(0,0,pos-r+1,pos-r+1)
  Mbottom=M.submatrix(pos+1,pos+1,s-pos-1,s-pos-1)
  return jordan_block(1,r).transpose().block_sum(Mtop).block_sum(Mbottom)    
###
# Jordan basis with vector u
# A is a matrix V is an invariant subspace killed by some (A-evalue*Id)^N 
# we assume that u can be the top vector of some Jordan block 
# the function returns a list which contains all top vectors in a Jordan base for V
# and u is the first vector in this list
#
# The function is recursive and so we need to include the case in which u=0.
# In this case, the function just returns a list of top vectors in a Jordan base for V.
# In particular, this function generalizes the standard algorithm to construct a Jordan basis
# 
# ***** Here is the only change wrt version 1.0: use of ImageUnderMatrix() function ***
#
def TopJordanBasisWithVector(u,A,V,evalue=1):
  if V.dimension()==0:
    return []
  E=parent(u); d=E.dimension(); K=E.base_field()
  Id=matrix(K,d,d,evalue)
  J=TopJordanBasisWithVector((A-Id)*u,A,ImageUnderMatrix(E,V,A-Id),evalue)
  bottomJ=[(A-Id)^(JordanHeight(J[i],A)-1)*J[i] for i in range(len(J))]
  if (A-Id)*u==0:
    if u==0:
      S=[]
    else:
      S=[u]
    H=extend(bottomJ+S,(A-Id).transpose().kernel().intersection(V),E)[len(bottomJ)+len(S):]
    L=[Hom(E,E)((A-Id).transpose()).preimage_representative(J[i]) for i in range(len(J))]
    return S+L+H
  else:
    H=extend(bottomJ,(A-Id).transpose().kernel().intersection(V),E)[len(bottomJ):]
    L=[Hom(E,E)((A-Id).transpose()).preimage_representative(J[i]) for i in [1..len(J)-1]]
    return [u]+L+H
###
# A is a matrix and u is a vector killed by some (A-evalue*Id)^N.
# The previous function returns the top vectors in a Jordan basis
# and this function completes those top vectors to a Jordan basis
# of the kernel of (A-evalue*Id)^N followed by a standard basis of the 
# complementary invariant subspace.
# The output is given as a change of basis matrix.
#
def JordanBasisWithVector(u,A, evalue=1):
  E=parent(u); d=E.dimension(); K=E.base_field() 
  Id=matrix(K,d,d,evalue)
  Poly.<z>=(E.base_field())[]
  p=A.minimal_polynomial('z')
  N=max([i for i in range(d+1) if ((z-evalue)^i).divides(p)])
  q=p//(z-evalue)^N
  V=((A-Id)^N).right_kernel()
  W=(q(A)).right_kernel()
  J=TopJordanBasisWithVector(u,A,V)
  L=[]
  for i in range(len(J)):
    L=L+[(A-Id)^j*J[i] for j in range(JordanHeight(J[i],A,evalue))]
  L=L+W.basis()
  return matrix(L).transpose()
###
# M is a matrix and this function conjugates M to a "real normal form"
# This normal form is a s follows. We find all real eigenvalues of M
# and we get a Jordan basis for the invariant subspace corresponding to all these 
# real eigenvalues.
# Then, we consider the complementary invariant subspace corresponding to
# the irreducible quadratic factors of the characteristic polynomial
# and we get a rational basis for this subspace.
# Then, we join the Jordan basis and the rational basis to get a basis of
# the full vector space.
#
# We need to extend M to the real algebraic Field AA. 
# If the matrix M is large or has a lot of non-rational eigenvalues, the computation
# of this function can be very demanding because we need to use the functions 
# jordan_form() and is_similar() in the field AA
#
def real_form(M,transformation=False):
  K=parent(M).base_ring()
  N=M.change_ring(QQbar)
  Ev=N.eigenvalues()
  REv=[Ev[i] for i in range(len(Ev)) if Ev[i] in AA]
  QEv=[Ev[i] for i in range(len(Ev)) if Ev[i] in K]
  if len(Ev)==len(REv):
    if len(Ev)==len(QEv):
      if not transformation:
        return M.transpose().jordan_form().transpose()
      MJ,PJ=M.transpose().jordan_form(transformation=True)
      return MJ.transpose(),(PJ^(-1)).transpose()
    if not transformation:
      return M.transpose().change_ring(AA).jordan_form()
    MJ,MP=M.transpose().change_ring(AA).jordan_form(transformation=True)
    return MJ.transpose(),(PJ^(-1)).transpose()
  if len(REv)==0:
    q=identity_matrix(M.nrows())
  else:
    q=prod([N-REv[i]*identity_matrix(M.nrows()) for i in range(len(REv))])
  h=prod([N-Ev[i]*identity_matrix(M.nrows()) for i in range(len(Ev)) if not Ev[i] in REv])
  RV=q.right_kernel(); Rd=RV.dimension()
  CV=h.right_kernel(); Cd=CV.dimension()
  P=matrix(RV.basis()+CV.basis()).transpose()
  X=P^(-1)*M*P
  MR=X.submatrix(0,0,Rd,Rd).change_ring(AA)
  MC=X.submatrix(Rd,Rd,M.nrows()-Rd,M.nrows()-Rd).change_ring(AA)
  JR,QR=MR.jordan_form(transformation=True)
  Basis1=matrix(QR).stack(matrix(QQ,Cd,Rd,0))
  JC=MC.rational_form()
  a,QC=MC.change_ring(QQbar).is_similar(JC.change_ring(QQbar),transformation=True)
  Basis2=matrix(QQ,Rd,Cd,0).stack((matrix(QC)^(-1)).change_ring(AA))
  MatriX=Basis1.augment(Basis2)
  if not transformation:
    return (MatriX^(-1)*X*MatriX).change_ring(AA)
  return (MatriX^(-1)*X*MatriX).change_ring(AA),P*MatriX
###
# f is an affinity and this function returns a normal form for it
# and, if required, and affinity h such that h^(-1)*f*h is in normal form.
# Please read the doc to learn about the normal form of an affinity
# and the way to get it.
# 
def NormalFormOfAffinity(f,transformation=False):
  E=parent(f(0)).change_ring(AA); d=E.dimension()
  Id=identity_matrix(d)
  fbar=LinearPartOfAffinity(f)
  A=fbar.matrix().transpose()
  rho=InvarianceLevel(f)
  if not transformation:
    B=real_form(A)
    if rho==0:
      return B.block_sum(identity_matrix(1))
    col=matrix(1,1,1).stack(matrix(d-1,1,0))
    P=raise_to_top(B,rho).augment(col)
    return P.stack(matrix(1,d,0).augment(matrix(1,1,1)))
  v=((A-Id)^(rho+1)).solve_right(((A-Id)^rho)*(f(0)))
  T=Affinity(E,-v,Hom(E,E)(Id))
  fprime=Conjugate(f,T)
  u=fprime(0)
  P=JordanBasisWithVector(u,A)
  g=Affinity(E,E(0),Hom(E,E)(P.transpose()))
  if rho==0:
    epsilon=0
  else:
    epsilon=1
  col=matrix(1,1,epsilon).stack(matrix(d-1,1,0))
  Q=P^(-1)*A*P
  X=Q.augment(col).stack(matrix(1,d,0).augment(matrix(1,1,1)))
  Poly.<z>=(E.base_field())[]
  p=A.minimal_polynomial('z')
  N=max([i for i in range(d+1) if ((z-1)^i).divides(p)])
  q=p//(z-1)^N
  dd=(q(Q)).right_kernel().dimension()
  L=Q.submatrix(Q.nrows()-dd,Q.nrows()-dd,dd,dd)
  R,Z=real_form(L,transformation=True)
  Y=identity_matrix(A.nrows()-dd).block_sum(Z)
  h=Affinity(E,E(0),Hom(E,E)(Y.transpose()))
  U=Y.block_sum(identity_matrix(1))
  return U^(-1)*X*U,CompositionOfAffinities(T,CompositionOfAffinities(g,h))
###
# Using the functions above, this function decides if two affinities are similar or not.
# with the argument transformation=True, it returns also an affinity
# h such that h^(-1)*f*h=g.
# Notice that if transformation is not required, similarity of the linear parts
# of f and g is decided comparing the rational forms, which is much faster than
# using sage's is_similar().
#
def IsSimilar(f,g,transformation=False):
  F=LinearPartOfAffinity(f).matrix().rational_form()
  G=LinearPartOfAffinity(g).matrix().rational_form()
  similar=(F==G and InvarianceLevel(f)==InvarianceLevel(g))
  if not similar or not transformation:
    return similar
  a,P=NormalFormOfAffinity(f,transformation=True)
  a,Q=NormalFormOfAffinity(g,transformation=True)
  return True, CompositionOfAffinities(P,InverseOfAffinity(Q))
###
# We give names to the affinities in dimension 1 and 2
#
def Name(f):
  E=parent(f(0)); d=E.dimension()
  fbar=LinearPartOfAffinity(f)
  q=fbar.characteristic_polynomial()
  if d>=4:
    raise NameError("this function is only for affinities in dimensions 1,2")
  if d==1:
    if InvarianceLevel(f)==1:
      return "Translation"
    if fbar.matrix()==identity_matrix(1):
      return "Identity"
    return "Homothecy"
  delta=q.discriminant()
  if delta<0:
    return "Elliptic"
  if delta==0:
    if q(1)==0:
      if fbar.matrix()==identity_matrix(2):
        if InvarianceLevel(f)==0:
          return "Identity"
        return "Translation"
      if InvarianceLevel(f)==0:
        return "Special Homology"
      return "Special Homology followed by Translation"
    if q==fbar.minimal_polynomial():
      return "Parabolic"
    return "Homothecy"      
  if delta>0:
    if not q(1)==0:
      return "Hyperbolic"
    if InvarianceLevel(f)==0:
      return "General Homology"
    return "General Homology followed by Translation"
#
###
### Functions for Chapter 5 in the book:
### Euclidean Affine Spaces
###
###
# Deciding if two varieties are orthogonal
#
def AreOrthogonal(X,Y):
  S=Direction(X).basis(); T=Direction(Y).basis()
  for s in S:
    for t in T:
      if not s.inner_product(t)==0:
        return False
  return True
###
# internal function to decide if X is a point or a variety
#
def is_point(X):
  x=X[0]
  try:
    a=len(x)
    return False
  except:
    return True
###
# internal function to decompose e\in F+G as e=u+v with u\in F, v\in G
# the function returns u
#
def decompose(e,F,G):
  if F.is_subspace(G):
    return parent(e)(0)
  if G.is_subspace(F):
    return e
  if not e in F+G:
    raise NameError('vector must be in the sum of the two subspaces')
  M=matrix(F.basis()).stack(matrix(G.basis()))
  X=M.solve_left(matrix(e))
  return sum([X[0,i]*F.basis()[i] for i in range(F.dimension())])
###
# internal function to return the orthogonal complement of a subspace V
#
def orthogonal_complement(E,V):
  e=E.basis()
  if V.dimension()==0:
    return E
  if E==V:
    return E.subspace([])
  metric=matrix([[e[i].inner_product(e[j]) for i in range(len(e))] for j in range(len(e))])
  M=matrix(V.basis())*metric
  X=M.right_kernel().basis()
  return E.subspace(X)
###
# internal function to return the orthogonal projection of e to V
#
def orthogonal_projection(e,V):
  return decompose(e,V,orthogonal_complement(parent(e),V))
###
# This function computes d=square of the distance between two varieties (or points) X,Y
# with points=True, the function returns d,P,Q with P\in X, Q\in Y and 
# d(P,Q)^2=d 
# 
# These computations are done using the results in sections 5.3 and 5.4 of the book.
#
def SquareDistance(X,Y,points=False):
  if is_point(X):
    XX=PointAsVariety(X)
  else:
    XX=X
  if is_point(Y):
    YY=PointAsVariety(Y)
  else:
    YY=Y
  F=Direction(XX); G=Direction(YY)
  PQ=Arrow(XX[0],YY[0])
  v=PQ-orthogonal_projection(PQ,F+G)
  d=v.inner_product(v)
  if not points:
    return d
  u1=decompose(PQ-v,F,G)
  u2=PQ-v-u1
  return d,XX[0]+u1,YY[0]-u2
###
# This function computes the variety orthogonal to X through the point P
#
def OrthogonalVariety(P,X):
  return Variety(P,orthogonal_complement(parent(P),Direction(X)))
###
### Chapter 6: Euclidean motions
###
###
# an internal function to get the metric of E
#
def metric(E):
  B=E.basis(); d=E.dimension()
  M=matrix([[B[i].inner_product(B[j]) for i in range(d)] for j in range(d)])
  return M
###
# an internal function to construct a diagonal matrix
#
def diagonal_matrix(S):
  d=len(S)
  A=matrix(1,1,S[0])
  for i in [1..d-1]:
    A=A.block_sum(matrix(1,1,S[i]))
  return A
###
# Function to decide if an affinity is an euclidean motion.
# It returns also a sequence of subspaces corresponding to 
# the eigenvalues +1, -1, and rotations, and
# an orthonormal basis in which the affinity is in normal form
#
def IsEuclideanMotion(f,subspaces=False,orthobase=False,):
  Q=f(0); E=parent(Q); d=E.dimension()
  Id=identity_matrix(d)
  M=metric(E)
  h=LinearPartOfAffinity(f)
  A=h.matrix().transpose()
  if not orthobase and not subspaces:
    return A.transpose()*M*A==Id
  if not A.transpose()*M*A==Id:
    return False
  p=A.minimal_polynomial()
  V0=(A-Id).right_kernel()
  V1=(A+Id).right_kernel()
  subs=[V0,V1]
  S=[q[0] for q in p.factor() if not q[0](1)==0 and not q[0](-1)==0]
  for s in S:
    subs.append(s(A).right_kernel())
  if subspaces and not orthobase:
    return True,subs
  base=matrix(QQ,1,d,0)
  for V in subs:
    if not V.dimension()==0:
      if V==V0 and not GlideVector(f)==0:
        mat=matrix(extend([GlideVector(f)],V,E)).gram_schmidt()[0]
      else:
        mat=matrix(V.basis()).gram_schmidt()[0]
      base=base.stack(mat)
  M=base.submatrix(1,0,d,d).transpose()
  D=diagonal_matrix([AA(sqrt((M.transpose()*M)[i,i]))^(-1) for i in range(d)])
  if subspaces:
    return True, subs,M*D
  return True,M*D
###
# We compute the glide vector of an euclidean motion
#
def GlideVector(f):
  Q=f(0); E=parent(Q); d=E.dimension()
  if not IsEuclideanMotion(f):
    raise NameError("Affinity must be an euclidean motion")
  boo,subs=IsEuclideanMotion(f,subspaces=True)
  V0=subs[0]
  if V0.dimension()==0:
    return E(0)
  V0perp=orthogonal_complement(E,V0)
  return E(decompose(Q,V0,V0perp))
###
# Extend to a basis
# B is a set of linearly independent vectors in V.
# V is a subspace of E.
# This function returns a basis of V which starts with B.
#
def extend(B,V,E):
  r=len(B)
  if r==0:
    return [E(V.basis()[i]) for i in range(V.dimension())]
  Bprime=[B[i] for i in range(r) if B[i] in V]
  Vprime=E.subspace(B)
  if not B==Bprime:
    raise NameError("vectors in partial basis must be in subspace")
  if not Vprime.dimension()==r:
    raise NameError("vectors must be linearly independent")
  X=copy.copy(B)
  while len(X)<V.dimension():
    for i in range(V.dimension()):
      if not V.basis()[i] in E.subspace(X):
        X.append(E(V.basis()[i]))
        break 
  return X
###
# Decide if two euclidean motions f,g are euclidian similar.
# If the optional parameter transformation is set to True,
# then it also returns an euclidean motion which conjugates 
# f to g.
#
def IsEuclideanSimilar(f,g,transformation=False):
  E=parent(f(0)); d=E.dimension()
  pf=LinearPartOfAffinity(f).matrix().characteristic_polynomial()
  pg=LinearPartOfAffinity(g).matrix().characteristic_polynomial()
  if not (IsEuclideanMotion(f) and IsEuclideanMotion(g)):
    raise NameError("Both affinities must be euclidean motions")
  vf=GlideVector(f)
  vg=GlideVector(g)
  sim=(pf==pg and vf.inner_product(vf)==vg.inner_product(vg))
  if not transformation:
    return sim
  if not sim:
    return False,False
  U,Hf=EuclideanNormalFormOfAffinity(f,transformation=True)
  U,Hg=EuclideanNormalFormOfAffinity(g,transformation=True)
  return True, MatrixAsAffinity(Hf*Hg^(-1),E.change_ring(AA))
###
# Orthogonal symmetry
#
def OrthogonalSymmetry(X):
  Y=OrthogonalVariety(BasePoint(X),X)
  V=Direction(Y)
  return Symmetry(X,V)
###
# Orthogonal projection
#
def OrthogonalProjection(X):
  E=parent(BasePoint(X))
  G=orthogonal_complement(E,Direction(X))
  return AffineProjection(X,G)
###
# Find the normal form of an euclidean motion
# and the transformation to it
#
def EuclideanNormalFormOfAffinity(f,transformation=False):
  E=parent(f(0)); d=E.dimension()
  pf=LinearPartOfAffinity(f).matrix().characteristic_polynomial()
  if not IsEuclideanMotion(f):
    raise NameError("Affinity is not an euclidean motion")
  vf=GlideVector(f)
  boo,Pf=IsEuclideanMotion(f,orthobase=True)
  Pff=Pf.block_sum(matrix(1,1,1))
  if not vf==0:
    P=BasePoint(FixedPoints(CompositionOfAffinities(f,AffineTranslation(-vf))))
    Hf=Pf.augment(matrix(P).transpose()).stack(matrix(1,d,0).augment(matrix(1,1,1)))
    if not transformation:
      return Hf^(-1)*AffinityAsMatrix(f)*Hf
    return Hf^(-1)*AffinityAsMatrix(f)*Hf,Hf
  ef=BasePoint(FixedPoints(f))
  Tf=AffinityAsMatrix(AffineTranslation(ef))
  if not transformation:
    return (Tf*Pff)^(-1)*AffinityAsMatrix(f)*Tf*Pff
  return (Tf*Pff)^(-1)*AffinityAsMatrix(f)*Tf*Pff,Tf*Pff
###
# Rotations
#
def RotationMatrix(P,theta):
  a=AA(cos(theta)); b=AA(sin(theta))
  x=b*P[1]+(1-a)*P[0]
  y=(1-a)*P[1]-b*P[0]
  M=matrix([[a,-b,x],[b,a,y],[0,0,1]])
  return M
###
# Function to identify a plane rotation.
# If it is a rotation, then as a second value it returns the fixed point 
# and a third value it returns the the angle of rotation
# in the form k*pi/n, as an element in the symbolic ring 
#
def IsRotation(M,params=False):
  if not M.nrows()==3:
    raise NameError('matrix must be a plane affinity')
  a=M[0,0]; b=M[0,1]; c=M[1,0]; d=M[1,1]
  rot=(a==d and b==-c and a^2+b^2==1 and not b==0)
  Q=(M-identity_matrix(3)).right_kernel().basis()[0]
  P=[Q[0]/Q[2],Q[1]/Q[2]]
  rotation=(rot and M[2,0]==0 and M[2,1]==0 and M[2,2]==1)
  if not params:
    return rotation
  if not rotation:
    return False,False,False
  for k in [0..25]:
    for n in [1..25]:
      if M==RotationMatrix(P,k*pi/n):
	return True,P,k*pi/n
  raise NameError('cannot find rotation angle')
      
  
    
  
    
    
    
 
    
    
    
    
    
    
    
    
    
    
    
    
    
    
    
    
    
    
    
    


